/*
 * des - DES encryption/decryption filter
 *
 *	Written by Koichiro Mori
 */

#include <stdio.h>
#include <string.h>
#if !unix
#include <stdlib.h>
#endif /* !unix */
#include <time.h>
#include <fcntl.h>


#if MSC
#include <io.h>
#endif

#ifdef __LIBC__
#include <unistd.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif /* O_BINARY */


extern char *getpass();

#include "des_crypt.h"


typedef char bool;
#define	YES	1
#define	NO	0
typedef unsigned char uchar;
#ifndef __LIBC__
typedef unsigned long ulong;
#endif

static char *Progname = "des";


/* Print usage */
void usage()
{
#ifdef HUMAN			/* Human Extension by WAKATONO at 26 Jun 1993 */
#ifdef __LIBC__
	fprintf(stderr, "DES (en|de)cryption filter by MORI Koichiro, ver 1.00\n");
	fprintf(stderr, "for X680x0( X6_01 ) by WAKATONO and PD libc version 1.1 by Mipro.\n");
#else
	fprintf(stderr, "DES (en|de)cryption filter by MORI Koichiro, ver 1.00( for X680x0( X6_01 ) by WAKATONO ) \n");
#endif
#else					/* End of Human Extension */
	fprintf(stderr, "DES (en|de)cryption filter by MORI Koichiro, ver 1.00\n");
#endif					/* Added this line by WAKATONO at 26 Jun 1993 */
	fprintf(stderr, "Usage: %s -e|-d [-b] [-k key] [input] [output]\n",
		Progname);
}


/* Print error message and die */
void error(msg, str)
	char *msg;
	char *str;
{
	fprintf(stderr, "%s: ", Progname);
	if (*msg == '+') {
		perror(str);
	} else {
		if (str)
			fprintf(stderr, "%s: ", str);
		fprintf(stderr, "%s\n", msg);
	}
	exit(1);
}


/* Get key from user */
void get_key(key)
	char *key;
{
	strncpy(key, getpass("Enter Key:"), 8);
}


/* Fill random 'n' byte sequence at address 'block' */
void set_random(block, n)
	char *block;
	int n;
{
	unsigned long now = time(NULL);
	unsigned long x;

	for (x = now; n; n--) {
		*block++ = x;
		if ((x >>= 8) == 0)
			x = now;
	}
}

/* Rubout key argument to prevent being peeped by ps(1) */
void killarg(arg)
	char *arg;
{
	while (*arg)
		*arg++ = 'X';
}

/* Entry point */
int main(argc, argv)
	int argc;
	char **argv;
{
	unsigned mode = -1;
	int ecb_flag = 0;
	int key_given = 0;
	int n, left;
	char key[8], ivec[8];
#define BLKSIZE 8192
	static char buff[2*BLKSIZE];
	char *filename, *p;
	int ifd, ofd;

	memset(key, 0, 8);
	while (++argv, --argc > 0 && *argv && **argv == '-' && argv[0][1]) {
		for (p = *argv + 1; *p; p++) {
			switch (*p) {
			case 'e':
				mode = DES_ENCRYPT;
				break;
			case 'd':
				mode = DES_DECRYPT;
				break;
			case 'b':
				ecb_flag = 1;
				break;
			case 'k': /* read key */
				if (--argc <= 0 || *++argv == NULL) {
					fprintf(stderr, "Missing key after -k\n");
					exit(1);
				}
				strncpy(key, *argv, 8);
				killarg(*argv);
				key_given = 1;
				break;
			case 'f': /* -f and -s options are ignored */
			case 's': /* Because we have no DES chip */
				break;
			default:
				fprintf(stderr, "Bad option: %s\n", *argv);
				usage();
				exit(1);
			}
		}
	}
	if (argc > 2) {
		usage();
		exit(1);
	}
	if (mode == -1) {
		fprintf(stderr, "No -e or -d flag\n");
		usage();
		exit(1);
	}
		
	if (argc >= 1 && strcmp(*argv, "-") != 0) {
		filename = *argv;
		if ((ifd = open(filename, O_RDONLY | O_BINARY)) < 0)
			error("+", filename);
	} else {
		filename = NULL;
		ifd = 0;
#if MSC
		setmode(0, O_BINARY);
#endif /* MSC */
	}
	if (argc >= 2 && strcmp(argv[1], "-") != 0) {
		if ((ofd = open(argv[1], O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666)) < 0)
			error("+", argv[1]);
	} else {
		ofd = 1;
#if MSC
		setmode(1, O_BINARY);
#endif /* MSC */
	}

	if (!key_given)
		get_key(key);

	des_setparity(key);
	memset(ivec, 0, 8);
	left = 0;
	while ((n = read(ifd, buff + left, sizeof(buff) - left)) > 0) {
		left += n;
		if (left > BLKSIZE) {
			if (ecb_flag)
				ecb_crypt(key, buff, BLKSIZE, mode);
			else
				cbc_crypt(key, buff, BLKSIZE, mode, ivec);
			write(ofd, buff, BLKSIZE);
			left -= BLKSIZE;
			memcpy(buff, buff + BLKSIZE, left);
		}
	}
	/* last block */
	n = (left & ~7) + 8;
	set_random(buff + left, n - left);
	buff[n - 1] = left & 7;
	if (ecb_flag)
		ecb_crypt(key, buff, n, mode);
	else
		cbc_crypt(key, buff, n, mode, ivec);
	if ((mode & DES_DIRMASK) == DES_DECRYPT) {
		if (left < 8 || (left & 7) >= 8 || buff[left - 1] >= 8) {
			error("data corrupted or invalid key", filename);
			exit(1);
		}
		n = (left - 8) + buff[left - 1];
	}
	write(ofd, buff, n);

	close(ofd);
	close(ifd);
	exit(0);
}
